;************************************************************************************************** 
;
;	Drehencoder nach BCD und Siebensegment
;
;	
;   Version 1.0	
;
;		- umschaltbar fr Drehencoder mit 1:1 bzw. 1:2 Impuls/Rastungen-Verhltnis	
;		- Umschaltung ber Jumper JP2 (Pin11) offen:  1:1  geschlossen:  1:2
;		- Ausgabe fr Siebensegmentanzeige (Pin12-18) wahlweise gemeinsame Katode/Anode
;		- Umschaltung ber Jumper JP1 (PIN19) offen:Katode geschlossen: Anode
;		- binre Ausgabe von 4 bit (Pin-9)
;		- getestet mit ALPS Drehencodern STEC12E07 (24/24) und STEC11B02 (15/30) 

;	Programmiert von Uwe Gubsch      uwe@gubsch.de
;	fr Wolfgang Harwardt alias Buebchen
;	an Leiterplattenlayout angepasste Belegung der Siebensegmentausgnge
;
;	Drehencoderauswertung frei nach P. Dannegger www.mikrocontroller.net
;   
;************************************************************************************************** 

.include "tn2313def.inc"

.equ Geber_A_Port	= PIND				; Port Drehgeber
.equ Geber_B_Port	= PIND
.equ Geber_A_Pin	= 0					; Anschlusspin A Drehgeber 
.equ Geber_B_Pin	= 1					; Anschlusspin B Drehgeber
.equ Entprellzeit	= 20				; Entprellzeit in ms fr Tasten


; Hier smtliche Register und deren Verwendung

;					r0					  reserviert fr Tabelle auslesen in Siebensegmentausgabe
.def zero  		= 	r1					; woimmer Null gebraucht wird
.def Zaehler	=	r2					; Zaehler Drehgeber
.def Entprell	=	r3					; Zaehler fr Tastenentprellung
.def Tasteninc	=	r4					; Tastenincrement 
;					r5					
;					r6					
;					r7
;					r8
;					r9
;					r10
;					r11
;					r12
;					r13
;					r14
;					r15		
;				 	r16
.def temp1 		= 	r17					; allgemeines Arbeitsregister
.def Tasten		=	r18					; Register fr Entprellte Tastenzustnde
; Tastenzustnde
	.equ Taste_auf	=	0				; Taste Aufwrtszhlen
	.equ Taste_ab	=	1				; Taste Abwrtszhlen
	.equ JP_2_1_Enc =	3				; Jumper an CPU-PIN 11: setzen wenn Drehencoder mit einem
										; Rasten/Impuls-Verhltnis von 2:1 benutzt werden sollen
	.equ JP_Anode	=	4				; Jumper an CPU-PIN 19: setzen wenn Siebensegmentanzeigen 
										; mit gemeinsamer Anode benutzt werden sollen
	.equ Quittung	=	5				; wird 1 gesetzt wenn Tastendruck verarbeitet wurde
										; Bit wird bei nderung einer Taste automatisch im
										; Unterprogramm Tastenabfrage zurckgesetzt
.def Tasten_neu	=	r19 
.def DG1		=	r20					; Arbeitsregister Unterprogramm-Drehgeber
.def DG2		=	r21					; Arbeitsregister Unterprogramm-Drehgeber
.def DG3		=	r22					; Arbeitsregister Unterprogramm-Drehgeber
.def Schiebe	=	r23
.def Segment	=	r24
.def BCD		=	r25
;					r26					XL		
;					r27					XH
;					r28					YL		
;					r29					YH		
;					r30					ZL		Tabelle auslesen in Siebensegmentausgabe
;					r31					ZH		Tabelle auslesen in Siebensegmentausgabe


.org 0x0000
           	rjmp    init               			; Reset Handler
.org OVF0addr
           	rjmp    timer0_1ms	     			; Timer Overflow Handler


;**************************************************************************************************
;
; Stackpointer initialisieren
			
init:		ldi     temp1, LOW(RAMEND)  		; Stackpointer auf RAM-Ende setzen
           	out     SPL, temp1					; Stackpointer setzen

; Timer initialisieren   	       	

			ldi		temp1, 1<<WGM01				; CTC-Mode setzen
	        out		TCCR0A, temp1				; 
			ldi		temp1, 1<<CS00 | 1<<CS01	; Teiler 64
			out		TCCR0B, temp1
			ldi		temp1, 1<<TOIE0				; TOIE0: Interrupt bei Timer Overflow
			out		TIMSK, temp1

; Ports initialisieren
			; PORTB
			; Pin 0-6 Anoden der Siebensegmentanzeige, Pin7 unbenutzt

			ldi     temp1, 0b01111111			; Pin 0-6 auf Ausgang - Pin 7 auf Eingang
           	out		DDRB, temp1
			ldi		temp1, 0b10000000			; Pull-UP-Widerstand Pin 7 aktivieren	
			out		PORTB, temp1		
			
			; PORTD
			; Pin 0,1 Drehgeber, Pin 2-5 BCD-Ausgabe, Pin6 unbenutzt
			ldi		temp1, 0b00111100			; Pin 0,1,6 auf Eingang, Pin 2,3,4,5 auf Ausgang
			out		DDRD, temp1					; alle Pins PortD auf Eingang
			ldi		temp1, 0b11000011
			out		PORTD, temp1				; Pull-UP-Widerstnde der Eingnge aktivieren

			; PORTA
			; Port ist unbenutzt, Pins werden in einen definierten Zustand geschaltet
			out		DDRA, zero					; alle Pins  Port A auf Eingang
			ldi		temp1, 0xFF
			out		PORTA, temp1				; Pull-UP-Widerstnde aktivieren 
			
			
; Sonstige Initialisierungen

			ldi		Schiebe, 0b00000001			; Schieberegister fr Siebensegmentmultiplexer
 			clr		zero	          			; Zhler initialisieren
			clr		BCD
			clr		Zaehler
			clr		Tasten
			clr		Tasten_neu
			rcall	Init_Drehgeber				; Drehgeber initialisieren
			sei	 								; Interrupt erlauben          	
	
;**************************************************************************************************
;
;			Hauptprogramm
;
;			eigentlich nur Warten auf einen Interrupt
;
Main:		rjmp 	Main						

;**************************************************************************************************
;
; Interruptroutine
			
timer0_1ms:	
			ldi		temp1, 0xc2					; Timer neu laden
			out		TCNT0, temp1			
			
			rcall 	Drehgeber					; Drehgeber abfragen
			add		zaehler, DG1				; Drehgeber Rckgabewert zhlen
								
			rcall	Tastenabfrage				; Tasten abfragen
			rcall	Tastenauswertung			; Abfrage auswerten
			add		zaehler, Tasteninc			; Tasten Rckgabewert zhlen

			

			mov		BCD, Zaehler				; Zaehler zur Auswertung bernehmen
			sbrs	Tasten, JP_2_1_Enc			; Wenn JP2 gesetzt nchste Befehl berspringen
			lsr		BCD							; nur ntig bei 1:1 Impulse:Rastung-Verhltnis
			lsr		BCD							; noch mal durch 2 teilen
			andi	BCD, 0x0F					; die oberen 4 bit beseitigen
			
	
			
			rcall	Siebensegmentausgabe		; Binrzahl auf die Siebensegmentanzeige ausgeben
			lsl		BCD							; Binrzahl auf die richtige Position im Ausgabe-
			lsl		BCD							; register verschieben
			ori		BCD, 0b11000011				; Die anderen Bits setzen wegen Pull-up-Widerstnde
			out		PORTD, BCD					; Ausgabe Binrzahl
			reti								; Ende Interruproutine

;**************************************************************************************************
;
; 			Abfrage Tasten
;
Tastenabfrage:
			clr		temp1
			sbis	PINA, 1					; Taste auf abfragen
			sbr		temp1, 1<<Taste_auf			
			sbis	PINA, 0					; Taste ab abfragen
			sbr		temp1, 1<<Taste_ab		
			sbis	PIND, 6					; Jumper 2 abfragen
			sbr		temp1, 1<<JP_2_1_Enc	
			sbis	PINB, 7					; Jumper 1 abfragen
			sbr		temp1, 1<<JP_Anode		
			cp		Tasten_neu, temp1		; Schauen ob sich was gendert hat
			breq	_tasten1				; wenn nicht
			mov		Tasten_neu, temp1		; wenn ja, neuer Zustand speichern
			ldi		temp1, Entprellzeit		; Entprellzeit aktivieren
			mov		Entprell, temp1
			ret
_tasten1:	cp		Entprell, zero			; Abfrage ob noch in der Entprellzeit
			breq	_tasten2				; wenn nein direkt zurck
			dec 	Entprell				; wenn ja Enprellzhler -1
			brne	_tasten2				; jetzt zuende, wenn nein zurck
			mov		Tasten, Tasten_neu		; wenn ja: Eingabe gltig
_tasten2:	ret

;**************************************************************************************************
;
; 			Auswertung Tasten
;			Rckgabewert in Tasteninc	hoch: +2 oder +4	2 bei 1:2 Impuls/Rastungen-Verhltnis
;									  runter: -2 oder -4    4 bei 1:1 Impuls/Rastungen-Verhltnis
;								 keine Taste: 0	
Tastenauswertung:
			clr		Tasteninc
			sbrc	Tasten, Quittung		; testen ob Quittungsbit gesetzt	
			ret								; wenn gesetzt: zurck mit Tasteninc=0
			sbr		Tasten, 1<<Quittung		; Quittungsbit setzen
			mov		temp1, tasten
			andi	temp1, 1<<Taste_auf | 1<<Taste_ab
			breq	_auswert1				; wenn keine Taste gedrckt: zurck
			inc		Tasteninc				; Tasteninc=1
			inc		Tasteninc				; Tasteninc=2
			sbrs	Tasten, JP_2_1_Enc
			lsl		Tasteninc				; Wenn 1:1 Impulse pro Rastung dann noch Tasteninc*2
			sbrc	Tasten, Taste_auf		; Wenn Taste_auf nicht gedrckt: zurck			
			neg		Tasteninc				; dann Taste_ab gedrckt: negatives Tastenicrement
_auswert1:
			ret


;**************************************************************************************************
;
; 			Abfrage Drehgeber
;			Rckgabewert in DG1 rechts: +1
;								links:	-1
;								sonst:	0
;	
;								
;			benutzte Register DG1, DG2, DG3;
Drehgeber:
			clr 	DG1						
			sbic 	Geber_A_Port,Geber_A_Pin	;Umwandlung Graycode in Binrcode
			inc 	DG1							;wenn hherweriges Bit gesetzt ist - dann
			sbis 	Geber_B_Port,Geber_B_Pin	;niederwertiges Bit invertieren
			rjmp 	_Geber_1				
			ldi 	DG2, 0x03
			eor 	DG1, DG2					; hier invertieren

_Geber_1:
			sub 	DG1, DG3					; Vergleichen
			sbrs 	DG1, 0x00					; schauen ob sich was gendert hat
			rjmp	_Geber_null					; wenn nicht: zurck
			add 	DG3, DG1					; aktuellen Zustand speichern
			andi 	DG1, 0x02					; Richtung auswerten
			sbrc 	DG1,1
			rjmp 	_Geber_links
			ldi		DG1,1						; bei Rechtsdrehung +1 zurckgeben
			ret

_Geber_links:
			ldi		DG1,-1						; bei Linksdrehung -1 zurckgeben
			ret

_Geber_null:
			clr 	DG1							; wenn Drehgeber nicht bettigt wurde 
			ret									; wird  Null zurckgegeben


;**************************************************************************************************
;
; 			Initialisierung Drehgeber
								
;			benutzte Register DG2, DG3;

Init_Drehgeber:
			clr 	DG3						
			sbic 	Geber_A_Port,Geber_A_Pin	;Umwandlung Graycode in Binrcode
			inc 	DG3							;wenn hherweriges Bit gesetzt ist - dann
			sbis 	Geber_B_Port,Geber_B_Pin	;niederwertiges Bit invertieren
			ret				
			ldi 	DG2, 0x03
			eor 	DG3, DG2					; hier invertieren
			ret


;**************************************************************************************************
;
; 			Siebensegmentausgabe
;
;			Schiebe, Segment 

Siebensegmentausgabe:
			ldi		ZL, LOW(Codes*2)	; Pointer fr die Abfrage der Codetabelle setzen
			ldi		ZH, HIGH(Codes*2)	
			add		ZL, BCD				; Binrzahl dazuaddieren
			adc		ZH, zero			; 
			lpm							; Segmentcode aus Codetabelle holen
			mov		Segment, r0			
			and		Segment, schiebe	; aktuelles Segment maskieren
			sbrc	Tasten, JP_Anode	; Bei Anzeige mit gemeinsamer Anode invertieren 
			com		Segment				;
			ori		Segment, 0b10000000	; Bit 7 setzen um Pin-UP-Wiederstand nicht abzuschalten
			out		PORTB, segment		; Ausgabe des aktuellen Segmantes auf den PortB
			lsl		schiebe				; Nach links schieben fr das nchste Segment
			sbrc	schiebe, 7			; schauen ob schon 7x nach links geschoben
			ldi		schiebe, 0b00000001	; wenn ja wieder neu laden
			ret	


;			Codetabelle fr Siebensegmentanzeige
Codes:                               	; Die Codetabelle fr die Ziffern 0 bis 9
                                     	; sie regelt, welche Segmente fr eine bestimmte
                                     	; Ziffer eingeschaltet werden mssen
;			Pinbelegung speziell angepasst fr Leiterplattenlayout von Buebchen
;           Gehusepin 12 = PB0 = Segment d
;           Gehusepin 13 = PB1 = Segment e    
;           Gehusepin 14 = PB2 = Segment f     
;           Gehusepin 15 = PB3 = Segment g
;           Gehusepin 16 = PB4 = Segment a
;           Gehusepin 17 = PB5 = Segment b
;           Gehusepin 18 = PB6 = Segment c
;
;			       0b-cbagfed  0b-cbagfed                            	
           .db     0b01110111, 0b01100000      ; 0: a, b, c, d, e, f		1: b, c
           .db     0b00111011, 0b01111001      ; 2: a, b, d, e, g			3: a, b, c, d, g
           .db     0b01101100, 0b01011101      ; 4: b, c, f, g 				5: a, c, d, f, g
           .db     0b01011111, 0b01110000      ; 6: a, c, d, e, f, g 		7: a, b, c
           .db     0b01111111, 0b01111101      ; 8: a, b, c, d, e, f, g 	9: a, b, c, d, f, g
		   .db     0b01111110, 0b01001111	   ; A: a, b, c, e, f, g,		B: c, d, e, f, g,
		   .db     0b00010111, 0b01101011	   ; C:	a, d, e, f,				D: b, c, d, e, g,	
		   .db     0b00011111, 0b00011110	   ; E:	a, d, e, f, g,			F: a, e, f, g,
